"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var net_1 = require("net");
var crypto_1 = require("crypto");
var Client_1 = require("./Client");
var Message_1 = require("./Message");
var UniqueID_1 = require("../utils/UniqueID");
/**
 *
 */
var PING_STRING = 'PING\u0000';
/**
 *
 */
var PONG_STRING = 'PONG\u0000';
/**
 *
 */
var HANDSHAKE_EVENT = 'GET';
/**
 *
 */
var HANDSHAKE_VERSION = '13';
/**
 *
 */
var HANDSHAKE_VERSION_REGEX = new RegExp(/Sec-WebSocket-Version: (.*)/);
/**
 *
 */
var HANDSHAKE_KEY_REGEX = new RegExp(/Sec-WebSocket-Key: (.*)/);
/**
 *
 */
var HANDSHAKE_PRE_ACCEPT = 'HTTP/1.1 101 Web Socket Protocol Handshake\r\nUpgrade: WebSocket\r\nConnection: Upgrade\r\nSec-WebSocket-Accept: ';
/**
 *
 */
var HANDSHAKE_ACCEPT = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11';
/**
 *
 */
var HANDSHAKE_HASH = 'sha1';
/**
 *
 */
var HANDSHAKE_DIGEST = 'base64';
/**
 * @author Bram Willem van der Kuip <vdkuipb@gmail.com>
 */
var Server = /** @class */ (function () {
    /**
     * Creates an instance of Server.
     * @param {string} [ip='127.0.0.1']
     * @param {number} [port=3000]
     */
    function Server(ip, port) {
        if (ip === void 0) { ip = '127.0.0.1'; }
        if (port === void 0) { port = 3000; }
        this.ip = ip;
        this.port = port;
        this.clients = [];
        this.handshakeAfterAccept = "\r\nWebSocket-Origin: http://" + this.ip + "\r\nWebSocket-Location: ws://" + this.ip + ":" + this.port + "/\r\n\r\n";
    }
    /**
     * TODO: Add the clientError event.
     * Starts the instance of this Server.
     */
    Server.prototype.start = function () {
        this.server = net_1.createServer();
        this.server.listen(this.port, this.ip);
        this.boundOnError = this.onError.bind(this);
        this.boundOnConnection = this.onConnection.bind(this);
        this.server.on("error" /* ERROR */, this.boundOnError);
        this.server.on("connection" /* CONNECTION */, this.boundOnConnection);
    };
    /**
     *
     * @param {Error} error
     */
    Server.prototype.onError = function (error) {
        console.log(error);
    };
    /**
     *
     * @param {Error} error
     */
    Server.prototype.onClientError = function (client, error) {
        console.log(error);
    };
    /**
     * Called the first time a socket is connected to the server.
     * Lets the server listen once for the data event to handle the handshake.
     * @param {Socket} socket
     */
    Server.prototype.onConnection = function (socket) {
        socket.once("data" /* DATA */, this.acceptHandshake.bind(this, socket));
    };
    /**
     * TODO: If returns false close socket and remove it.
     * Handles the handshake of the socket and the server.
     * Return if the handshake went succesfully
     * @param {Socket} socket
     * @param {object} handshake
     * @returns {boolean}
     */
    Server.prototype.acceptHandshake = function (socket, handshakeObject) {
        if (handshakeObject === null)
            return;
        var handshake = handshakeObject.toString();
        // For healthdaemon
        if (handshake.substr(0, 5) === PING_STRING) {
            socket.write(PONG_STRING);
        }
        if (handshake.substr(0, 3) === HANDSHAKE_EVENT) {
            var version = handshake.match(HANDSHAKE_VERSION_REGEX);
            if (version && version[1] === HANDSHAKE_VERSION) {
                var key = handshake.match(HANDSHAKE_KEY_REGEX);
                if (!key) {
                    return false;
                }
                var acceptKey = crypto_1.createHash(HANDSHAKE_HASH)
                    .update(key[1] + HANDSHAKE_ACCEPT)
                    .digest(HANDSHAKE_DIGEST);
                socket.write(HANDSHAKE_PRE_ACCEPT +
                    acceptKey +
                    this.handshakeAfterAccept);
                this.onAcceptedHandshake(socket);
                return true;
            }
        }
        socket.destroy();
        return false;
    };
    /**
     * TODO: Maybe find another solution for this.
     * Handles the client events and is called if the handshake was succesfull.
     * @param {Socket} socket
     */
    Server.prototype.onAcceptedHandshake = function (socket) {
        var client = new Client_1.Client(socket, UniqueID_1.UniqueID.generate());
        client.boundOnSocketData = this.onData.bind(this, client);
        client.boundOnSocketClose = this.onClose.bind(this, client);
        client.boundOnSocketError = this.onClientError.bind(this, client);
        client.socket.on("data" /* DATA */, client.boundOnSocketData);
        client.socket.on("close" /* CLOSE */, client.boundOnSocketClose);
        client.socket.on("error" /* ERROR */, client.boundOnSocketError);
        this.clients.push(client);
        this.onClientConnection(client);
    };
    /**
     * TODO: Handle incorrect message format.
     * Called when a socket sends data to the server.
     * @param {Client} client
     * @param {object} data
     */
    Server.prototype.onData = function (client, data) {
        var message = Message_1.MessageFormatter.decrypt(data);
        switch (message.event) {
            case "close" /* CLOSE */:
                this.onClose(client);
                break;
            case "invalid_message" /* INVALID_MESSAGE */:
                this.onClose(client);
                break;
            case "pong" /* PONG */:
                client.keepAlive.finish();
                break;
            default:
                this.onClientData(client, message);
                break;
        }
    };
    /**
     * TODO: Check the bounds and removes.
     * Called when a socket closes.
     * @param {Client} client
     */
    Server.prototype.onClose = function (client) {
        this.clients.splice(this.clients.indexOf(client));
        this.onClientClose(client);
        client.destroyInstance();
    };
    /**
     * Sends a given message to the given client.
     * @param {Client} client
     * @param {Message} message
     */
    Server.prototype.send = function (client, message) {
        client.send(message);
    };
    /**
     * TODO: Now send message is doing the message to buffer for every client.
     * Sends a given message to all clients connected.
     * @param {Message} message
     */
    Server.prototype.broadcast = function (message) {
        for (var i = this.clients.length - 1; i >= 0; i -= 1) {
            this.clients[i].send(message);
        }
    };
    /**
     * Destroys this instance of Server.
     */
    Server.prototype.destroyInstance = function () {
        this.server.removeListener("error" /* ERROR */, this.boundOnError);
        this.server.removeListener("connection" /* CONNECTION */, this.boundOnConnection);
    };
    return Server;
}());
exports.Server = Server;
