import { Socket } from 'net';
import { Message, MessageFormatter } from './Message';
import { Destroyable } from '../utils/Destroyable';
import { SocketKeepAlive } from './SocketKeepAlive';

/**
 * A enum of all the net socket events.
 */
export const enum NetSocketEvents {
    ERROR = 'error',
    DATA = 'data',
    CLOSE = 'close',
}
/**
 *
 */
const SOCKET_WRITE_ENCODING: string = 'binary';

/**
 * TODO: More private.
 *
 * @author Bram Willem van der Kuip <vdkuipb@gmail.com>
 */
export class Client implements Destroyable {
    /**
     * The unique id of this client instance.
     */
    public id: string;
    /**
     * The socket that is listened to an writen data to for this client.
     */
    public socket: Socket;
    /**
     * Object that handles the ping/pong events.
     */
    public keepAlive: SocketKeepAlive;
    /**
     * Bound methods that are saved so they can be removed if this client is disconnected.
     *
     */
    public boundOnSocketData: (data: object) => any;
    public boundOnSocketClose: () => any;
    public boundOnSocketError: () => any;
    /**
     * Creates an instance of Client.
     * @param {Socket} socket The socket of this client.
     * @param {string} id The id of this client.
     */
    public constructor(socket: Socket, id: string) {
        this.socket = socket;
        this.id = id;
        this.keepAlive = new SocketKeepAlive(socket);
    }

    /**
     * Sends message to the socket of this client.
     * @param {Message} The message to send.
     */
    public send(message: Message): void {
        const data: any = MessageFormatter.encrypt(message);
        this.socket.write(data, SOCKET_WRITE_ENCODING);
    }

    /**
     * Destroys this instance of Client.
     */
    public destroyInstance(): void {
        this.keepAlive.destroyInstance();
        delete this.keepAlive;
        this.socket.removeListener(NetSocketEvents.CLOSE, this.boundOnSocketClose);
        this.socket.removeListener(NetSocketEvents.DATA, this.boundOnSocketData);
        this.socket.destroy();
        delete this.id;
        delete this.socket;
        delete this.boundOnSocketClose;
        delete this.boundOnSocketData;
    }

}
