import { Socket, Server as netServer, createServer } from 'net';
import { createHash } from 'crypto';

import { Client, NetSocketEvents } from './Client';
import { Message, MessageFormatter } from './Message';
import { Destroyable } from '../utils/Destroyable';
import { UniqueID } from '../utils/UniqueID';

/**
 * An enum of all the net server events.
 */
export const enum NetServerEvents {
    ERROR = 'error',
    CONNECTION = 'connection',
}
/**
 *
 */
export const enum BaseMessageEvents {
    CLOSE = 'close',
    INVALID_MESSAGE = 'invalid_message',
    PONG = 'pong',
    PING = 'ping',
}
/**
 *
 */
const PING_STRING: string = 'PING\u0000';
/**
 *
 */
const PONG_STRING: string = 'PONG\u0000';
/**
 *
 */
const HANDSHAKE_EVENT: string = 'GET';
/**
 *
 */
const HANDSHAKE_VERSION: string = '13';
/**
 *
 */
const HANDSHAKE_VERSION_REGEX: RegExp = new RegExp(/Sec-WebSocket-Version: (.*)/);
/**
 *
 */
const HANDSHAKE_KEY_REGEX: RegExp = new RegExp(/Sec-WebSocket-Key: (.*)/);
/**
 *
 */
const HANDSHAKE_PRE_ACCEPT: string = 'HTTP/1.1 101 Web Socket Protocol Handshake\r\nUpgrade: WebSocket\r\nConnection: Upgrade\r\nSec-WebSocket-Accept: ';
/**
 *
 */
const HANDSHAKE_ACCEPT: string = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11';
/**
 *
 */
const HANDSHAKE_HASH: string = 'sha1';
/**
 *
 */
const HANDSHAKE_DIGEST: any = 'base64';

/**
 * @author Bram Willem van der Kuip <vdkuipb@gmail.com>
 */
export abstract class Server implements Destroyable {
    /**
     * The ip of this server.
     */
    private ip: string;
    /**
     * The port this server listens to.
     */
    private port: number;
    /**
     * An array of al the connected clients.
     */
    private clients: Client[];
    /**
     * The net.server isntance.
     */
    private server: netServer;
    /**
     * Part of the handshake string that we create beforehand with the ip and port.
     */
    private handshakeAfterAccept: string;
    /**
     * Bound methods that are saved so they can be removed if this server is destroyed.
     */
    private boundOnError: (error: Error) => any;
    private boundOnConnection: (socket: Socket) => any;
    /**
     * Creates an instance of Server.
     * @param {string} [ip='127.0.0.1']
     * @param {number} [port=3000]
     */
    public constructor(ip: string = '127.0.0.1', port: number = 3000) {
        this.ip = ip;
        this.port = port;

        this.clients = [];
        this.handshakeAfterAccept = `\r\nWebSocket-Origin: http://${this.ip}\r\nWebSocket-Location: ws://${this.ip}:${this.port}/\r\n\r\n`;
    }

    /**
     * TODO: Add the clientError event.
     * Starts the instance of this Server.
     */
    public start(): void {
        this.server = createServer();
        this.server.listen(this.port, this.ip);

        this.boundOnError = this.onError.bind(this);
        this.boundOnConnection = this.onConnection.bind(this);
        this.server.on(NetServerEvents.ERROR, this.boundOnError);
        this.server.on(NetServerEvents.CONNECTION, this.boundOnConnection);
    }

    /**
     *
     * @param {Error} error
     */
    private onError(error: Error): void {
        console.log(error);
    }

    /**
     *
     * @param {Error} error
     */
    private onClientError(client: Client, error: Error): void {
        console.log(error);
    }

    /**
     * Called the first time a socket is connected to the server.
     * Lets the server listen once for the data event to handle the handshake.
     * @param {Socket} socket
     */
    private onConnection(socket: Socket): void {
        socket.once(NetSocketEvents.DATA, this.acceptHandshake.bind(this, socket));
    }

    /**
     *
     * @param {Client} client
     */
    protected abstract onClientConnection(client: Client): void;

    /**
     * TODO: If returns false close socket and remove it.
     * Handles the handshake of the socket and the server.
     * Return if the handshake went succesfully
     * @param {Socket} socket
     * @param {object} handshake
     * @returns {boolean}
     */
    private acceptHandshake(socket: Socket, handshakeObject: object): boolean {
        if (handshakeObject === null) return;
        const handshake = handshakeObject.toString();

        // For healthdaemon
        if (handshake.substr(0, 5) === PING_STRING) {
            socket.write(PONG_STRING);
        }

        if (handshake.substr(0, 3) === HANDSHAKE_EVENT) {
            const version = handshake.match(HANDSHAKE_VERSION_REGEX);
            if (version && version[1] === HANDSHAKE_VERSION) {
                const key = handshake.match(HANDSHAKE_KEY_REGEX);
                if (!key) {
                    return false;
                }

                const acceptKey = createHash(HANDSHAKE_HASH)
                .update(key[1] + HANDSHAKE_ACCEPT)
                .digest(HANDSHAKE_DIGEST);
                socket.write(HANDSHAKE_PRE_ACCEPT +
                    acceptKey +
                    this.handshakeAfterAccept);
                this.onAcceptedHandshake(socket);
                return true;
            }
        }

        socket.destroy();
        return false;
    }

    /**
     * TODO: Maybe find another solution for this.
     * Handles the client events and is called if the handshake was succesfull.
     * @param {Socket} socket
     */
    private onAcceptedHandshake(socket: Socket): void {
        const client = new Client(socket, UniqueID.generate());

        client.boundOnSocketData = this.onData.bind(this, client);
        client.boundOnSocketClose = this.onClose.bind(this, client);
        client.boundOnSocketError = this.onClientError.bind(this, client);

        client.socket.on(NetSocketEvents.DATA, client.boundOnSocketData);
        client.socket.on(NetSocketEvents.CLOSE, client.boundOnSocketClose);
        client.socket.on(NetSocketEvents.ERROR, client.boundOnSocketError);

        this.clients.push(client);
        this.onClientConnection(client);
    }

    /**
     * TODO: Handle incorrect message format.
     * Called when a socket sends data to the server.
     * @param {Client} client
     * @param {object} data
     */
    private onData(client: Client, data: object): void {
        const message: Message = MessageFormatter.decrypt(data);

        switch (message.event) {
        case BaseMessageEvents.CLOSE:
            this.onClose(client);
            break;
        case BaseMessageEvents.INVALID_MESSAGE:
            this.onClose(client);
            break;
        case BaseMessageEvents.PONG:
            client.keepAlive.finish();
            break;
        default:
            this.onClientData(client, message);
            break;
        }
    }

    /**
     *
     * @param {Client} client
     * @param {Message} message
     */
    protected abstract onClientData(client: Client, message: Message): void;

    /**
     * TODO: Check the bounds and removes.
     * Called when a socket closes.
     * @param {Client} client
     */
    private onClose(client: Client): void {
        this.clients.splice(this.clients.indexOf(client));
        this.onClientClose(client);
        client.destroyInstance();
    }

    /**
     *
     * @param {Client} client
     */
    protected abstract onClientClose(client: Client): void;

    /**
     * Sends a given message to the given client.
     * @param {Client} client
     * @param {Message} message
     */
    public send(client: Client, message: Message): void {
        client.send(message);
    }

    /**
     * TODO: Now send message is doing the message to buffer for every client.
     * Sends a given message to all clients connected.
     * @param {Message} message
     */
    public broadcast(message: Message): void {
        for (let i  = this.clients.length - 1; i >= 0; i -= 1) {
            this.clients[i].send(message);
        }
    }

    /**
     * Destroys this instance of Server.
     */
    public destroyInstance(): void {
        this.server.removeListener(NetServerEvents.ERROR, this.boundOnError);
        this.server.removeListener(NetServerEvents.CONNECTION, this.boundOnConnection);
    }

}
